/**
 * @file object.h
 * Header for Object class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _OBJECT_H
#define _OBJECT_H

#include "main.h"
#include "actor.h"
#include "utils.h"

#include <gfx/vec4.h>
#include <gfx/mat4.h>
#include <gfx/quat.h>

#include <ode/ode.h>

class Object;
class PhysicalEffect;
class Effect;
class Model;
class Camera;

/**
 * A method pointer for effect handlers (what gets called when the Object
 * recieves a certain effect).
 */
typedef void (Object::*EffectHandler)(Effect *);

/**
 * A physical, visible object.
 *
 * Objects are physical things in the scene, and interact with each other
 * via effects.  To receive effects, an Object must implement
 * get_effect_handler(), and return the function it wants called whenever
 * it receives a effect of a given type.
 */
class Object : public Actor {
public:

    /**
     * Constructor.
     *
     * @param model the model (graphical representation) of this object.
     * The object will not obtain possession of the model (and will not
     * delete it when destructed).
     * @param position 3d position of the object.
     * @param orientation 3d orientation of the object
     * @scale scaling factor of the object
     * @param mass the mass in kg.
     */
    Object( Model *model, Vec3 position = Vec3( 0.0, 0.0, 0.0 ), dMatrix3 orientation = NULL, double scale = 1.0, double mass = 1.0 );
    virtual ~Object( void );

    /**
     * Have the object determine what kind of Effect effect is, and return a
     * method pointer that will handle applications of that effect.
     *
     * @param effect the effect that we wish to find the handler for.
     * @return the handler (method pointer) for the given effect.
     */
    virtual EffectHandler get_effect_handler( Effect *effect );

    /**
     * Render the object to the screen.  I think this will actually just
     * call the model's render() function.
     */
    virtual void render( bool selected = false );

    virtual void tick( double t ) { }

    // keep the object's "sim" positions/orientations updated, and reset them
    // when necessary
    virtual void update_sim( void );
    virtual void reset_sim( void );

    //@{
    /**
     * Accessor functions.
     *
     * These give access to private object data.  See variables for
     * details.
     */
    void set_position( const Vec3& pos ) { dBodySetPosition( _body_id, pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void set_orientation_mat( const dMatrix3 orientation ) { dBodySetRotation( _body_id, orientation ); }
    void set_orientation_quat( const dQuaternion orientation ) { dBodySetQuaternion( _body_id, orientation ); }
    //void set_mass( double m ) { _mass = m; }
    void set_scale( const double s ) { _scale = s; }
    void set_stationary( const bool s ) { _stationary = s; }
    void set_ethereal( const bool e ) { _ethereal = e; }
    void set_on( const bool o ) { _on = o; }

    Vec3 get_position( void ) const { const float* pos = dBodyGetPosition( _body_id ); return Vec3( pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void get_orientation_mat( dMatrix3 dmat3 ) const { dRealPtrTodMatrix3( dmat3, dBodyGetRotation( _body_id ) ); }
    void get_orientation_quat( dQuaternion dquat ) const { dRealPtrTodQuaternion( dquat, dBodyGetQuaternion( _body_id ) ); }
    //virtual double get_mass( void ) { return _mass; }
    virtual double get_scale( void ) const { return _scale; }
    virtual bool is_stationary( void ) const { return _stationary; }
    virtual bool is_ethereal( void ) const { return _ethereal; }
    virtual bool is_on( void ) const { return _on; }

    Vec3 get_sim_position( void ) const { return _sim_position; }
    void get_sim_orientation_mat( dMatrix3 dmat3 ) const { dMatrix3TodMatrix3( dmat3, _sim_orientation ); }
    void get_sim_orientation_quat( dQuaternion dquat ) const { dRtoQ( _sim_orientation, dquat ); }
    bool is_sim_on( void ) const { return _sim_on; }

    const dBodyID get_body_id() { return _body_id; }
    const dGeomID get_geom_id() { return _geom_id; }
    //@}

    friend void mouse_drag_object( Camera *camera, Object *obj, Vec2 prev, Vec2 cur );

protected:

    Model* _model;              ///< my graphical representation
    double _scale;              ///< the scaling factor

    dBodyID _body_id;
    dGeomID _geom_id;

    Vec3 _sim_position;
    dMatrix3 _sim_orientation;
    bool _sim_on;

    // shouldn't need these anymore
#if 0
    Vec3 _position;             ///< 3d position
    Vec3 _velocity;             ///< 3d translational velocity
    Vec3 _net_force;            ///< total force accumulated this tick

    double _rotation_angle;     ///< angle to rotate
    Vec3 _rotation_axis;        ///< axis to rotate about
    Vec3 _ang_velocity;         ///< 3 components of angular velocity
    Vec3 _net_torque;           ///< total torque accumulated this tick

    double _mass;               ///< my mass
#endif

    bool _stationary;           ///< am i movable?
    bool _ethereal;             ///< can objects collide with me / can I overlap objects

    bool _on;                   ///< am i activated?
    // other variables will go here??

private:

    /**
     * Default handler for physical effects (ie, forces).
     *
     * Most objects will not have a specialized handler for Physical
     * effects, in which case those effects will be handled here, via
     * Newton's laws.
     * 
     * @param effect the effect that we're receiving.
     */
    void _handle_physical_effect( PhysicalEffect *effect );
};

#endif // header guard
