#include "level.h"
#include "model.h"
#include "msmodel.h"
#include "box.h"
#include "sphere.h"
#include "cylinder.h"

inline string get_string( const CL_DomElement& element, string name )
{
    return element.get_attribute( name );
}

inline float get_float( const CL_DomElement& element, string name )
{
    return CL_String::to_float( element.get_attribute( name ) );
}

inline double get_double( const CL_DomElement& element, string name )
{
    return CL_String::to_double( element.get_attribute( name ) );
}

inline bool get_bool( const CL_DomElement& element, string name )
{
    return CL_String::to_bool( element.get_attribute( name ) );
}

inline Vec3 get_vec3( const CL_DomElement& element, string name )
{
    vector<string> tokens =
        CL_String::tokenize( element.get_attribute( name ), " " );

    return Vec3( CL_String::to_double( tokens[ 0 ] ), 
            CL_String::to_double( tokens[ 1 ] ), 
            CL_String::to_double( tokens[ 2 ] ) );
}

inline CL_DomElement get_element( CL_ResourceManager manager, string resource )
{
    return manager.get_resource( resource ).get_element();
}


Object *new_object_from_dom( const CL_DomElement& element )
{
    Object *obj = NULL;

    // all objects need at least these attributes
    if ( !element.has_attribute( "type" )
            || !element.has_attribute( "position" )
            || !element.has_attribute( "forward" )
            || !element.has_attribute( "up" )
            || !element.has_attribute( "model" ) ) {
        return NULL;
    }

    string type = get_string( element, "type" );
    Vec3 position = get_vec3( element, "position" );
    Vec3 forward = get_vec3( element, "forward" );
    Vec3 up = get_vec3( element, "up" );
    string model_filename = get_string( element, "model" );

    // put these only in the objects that need them.
    double mass = get_double( element, "mass" );
    double scale = get_double( element, "scale" );
    bool on = get_bool( element, "on" );
    bool stationary = get_bool( element, "stationary" );
    Model *model = new MSModel( "resources/models/", model_filename );

    dMatrix3 orientation;
    Vec3 right = cross( forward, up );

    dRFrom2Axes( orientation, right[ 0 ], right[ 1 ], right[ 2 ], up[ 0 ], up[ 1 ], up[ 2 ] );

    if( CL_String::compare_nocase( type, "box" ) ) {
        obj = new Box( model, position, orientation, scale, mass );
    }
    else if( CL_String::compare_nocase( type, "sphere" ) ) {
        obj = new Sphere( model, position, orientation, scale, mass );
    }
    else if( CL_String::compare_nocase( type, "cylinder" ) ) {
        obj = new Cylinder( model, position, orientation, scale, mass );
    }
    else {
        cerr << "Object type not recognized" << endl;
        return NULL;
    }

    return obj;
}

Level::Level( string filename ) :
    _filename( filename )
{
    cout << "Loading: " << filename << endl;

    CL_ResourceManager level_resources( filename );

    _description = CL_String::load( "Description/LevelDescription", &level_resources );
    cout << "Description: " << _description << endl;

    _id = CL_Integer( "Description/ID", &level_resources );
    cout << "ID: " << _id << endl;

    _world_size = get_vec3( get_element( level_resources, "Description/worldSize" ), "value" );
    cout << "World Size: " << _world_size << endl;

    _camera_radius = get_double( get_element( level_resources, "Description/Camera" ), "radius" );
    _camera_elevation = get_double( get_element( level_resources, "Description/Camera" ), "elevation" );
    _camera_azimuth = get_double( get_element( level_resources, "Description/Camera" ), "azimuth" );
    cout << "Camera Radius: " << _camera_radius << " Camera Elevation: " << _camera_elevation << " Camera Azimuth: " << _camera_azimuth << endl;

    list<string> goal_list = level_resources.get_resources_of_type( "goal", "Goals" );
    list<string>::iterator itr;

    for( itr = goal_list.begin(); itr != goal_list.end(); ++itr ) {
        CL_Resource goal_resource = level_resources.get_resource( *itr );

        goal_resource.load();

        CL_DomElement element = goal_resource.get_element();

        int ID = CL_String::to_int( element.get_attribute( "ID" ) );

        string type = element.get_attribute( "type" );

        int object_id = CL_String::to_int( element.get_attribute( "objectID" ) );

        Vec3 position( get_vec3( element, "position" ) );

        float radius = CL_String::to_float( element.get_attribute( "radius" ) );

        GoalType goal_type;
        if( CL_String::compare_nocase( type, "inside" ) )
            goal_type = INSIDE;
        else if( CL_String::compare_nocase( type, "outside" ) )
            goal_type = OUTSIDE;
        else if( CL_String::compare_nocase( type, "activate" ) )
            goal_type = ACTIVATE;
        else if( CL_String::compare_nocase( type, "deactivate" ) )
            goal_type = DEACTIVATE;
        else
            cerr << "Goal not recognized" << endl;

        cerr << "adding goal " << ID << " with obj id " << object_id << endl;
        _goals[ ID ] = new Goal( ID, goal_type, object_id, position, radius );

        goal_resource.unload();
    }

    list<string> actor_list = level_resources.get_resources_of_type( "actor", "Actors" );

    for( itr = actor_list.begin(); itr != actor_list.end(); ++itr ) {
        CL_Resource actor_resource = level_resources.get_resource( *itr );

        actor_resource.load();

        CL_DomElement element = actor_resource.get_element();

        int ID = CL_String::to_int( element.get_attribute( "ID" ) );

        string type = element.get_attribute( "type" );

        /*
        if( CL_String::compare_nocase( type, "gravity" ) ) {
            actors[ ID ] = new Gravity( get_vec3( element, "force" ) );
        }
        else
            cerr << "Actor not recognized" << endl;
        */

        actor_resource.unload();
    }

    list<string> object_list = level_resources.get_resources_of_type( "object", "Objects" );

    for( itr = object_list.begin(); itr != object_list.end(); ++itr ) {
        CL_Resource object_resource = level_resources.get_resource( *itr );

        object_resource.load();

        CL_DomElement element = object_resource.get_element();

        int ID = CL_String::to_int( element.get_attribute( "ID" ) );

        Object *obj = new_object_from_dom( element );
        if ( obj ) {
            _objects[ ID ] = obj;
        }
        else {
            cerr << "Error parsing object element" << endl;
        }

        object_resource.unload();
    }
}

Level::~Level( void )
{
    // mem leak, arent deleting the models associated with the objects
    ObjectList::iterator itr;

    // make sure the objects are removed from ode before either
    // reloading them or loading new ones
    for( itr = _objects.begin(); itr != _objects.end(); ++itr )
    {
        delete itr->second;
    }

    // should probably be doing this for actors as well
}

const bool Level::level_completed( void )
{
    bool completed = true;

    GoalIterator itr;

    for( itr = _goals.begin(); itr != _goals.end(); ++itr ) {
        Goal* goal = itr->second;

        // goals only have an object_id, not any more info about objects, so
        // we need to pass in the pointer to the object mentioned in the goal
        if( !goal->check_goal( _objects[ goal->get_object_id() ] ) ) {
            completed = false;
        }
    }


    return completed;
}

void Level::reset_sim( void )
{
    ObjectList::iterator itr;

    // reset the object's position and orientation to the position and
    // orientation it had before the simulation started
    for( itr = _objects.begin(); itr != _objects.end(); ++itr )
    {
        itr->second->reset_sim();
    }
}

void Level::reset_level( void )
{
    // do any other initialization needed
}
