#include "camera.h"

#include <gfx/gl.h>

#include <math.h>

Camera::Camera(ProjectType pt) :
    _radius(100.0),
    _elevation(0.0),
    _azimuth(0.0),
    _project_type(pt)
{
}

// y = r*sin(_elevation)
// x = r*cos(_elevation)*sin(_azimuth)
// z = r*cos(_elevation)*cos(_azimuth)
Vec3 Camera::get_position() const
{
    double rad_e = DEG_TO_RAD( _elevation );
    double rad_a = DEG_TO_RAD( _azimuth );
    double r_cos_e = _radius * cos( rad_e );

    return Vec3( r_cos_e * sin( rad_a ),
            _radius * sin( rad_e ),
            r_cos_e * cos( rad_a ) );
}

Vec3 Camera::get_right_vec()
{
    Vec3 right( get_forward_vec() ^ get_up_vec() );
    unitize( right );
    return right;
}

Vec3 Camera::get_up_vec()
{
    return Vec3( 0, 1, 0 );
}

Vec3 Camera::get_forward_vec()
{
    Vec3 forward( Vec3( 0 ) - get_position() );
    unitize( forward );
    return forward;
}

void Camera::apply_projection()
{
    GLuint w = CL_Display::get_width();
    GLuint h = CL_Display::get_height();

    glViewport( 0, 0, w, h );

    glMatrixMode( GL_PROJECTION );
    switch ( _project_type ) {
    case ORTHOGRAPHIC:
        gluOrtho2D( 0.0, w, 0.0, h );
        break;
    case PERSPECTIVE:
        // FIXME: parameterize, and pick better defaults
        gluPerspective( 45.0f, GLfloat( w ) / GLfloat( h ), 5.0f, 2000.0f );
        break;
    default:
        glLoadIdentity();
        break;
    }
}

void Camera::apply_view()
{
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glTranslated(0.0, 0.0, -_radius);
    glRotated(_elevation, 1.0, 0.0, 0.0);
    glRotated(-_azimuth, 0.0, 1.0, 0.0);

    // modelview = T*R*R.  first rotate around y axis, then about x axis,
    // then move camera straight backwards.  voila, spherical coordinates.
    // up: (0, 1, 0)
    // forward: (0, 0, 0) - position
}

std::ostream& operator<<(std::ostream & os, const Camera & camera)
{
    os << "radius: " << camera._radius
        << ", angle (" << camera._azimuth
        << ", " << camera._elevation << endl;
    os << "position: " << camera.get_position() << endl;

    return os;
}

#if 0
Camera::Camera( Vec3f pos, Vec3f forwardVec, Vec3f upVec,
        float fov, float nearPlane, float farPlane,
        int screenWidth, int screenHeight,
        ProjType projType )
{
    this->pos = pos;
    this->forwardVec = forwardVec;
    this->upVec = upVec;

    this->rotateSpeed = 0.05f;
    this->strafeSpeed = 2.5f;
    this->moveSpeed   = 2.5f;

    this->projType = projType;

    this->fov = fov;

    this->nearPlane = nearPlane;
    this->farPlane = farPlane;

    this->screenWidth = screenWidth;
    this->screenHeight = screenHeight;

    this->aspectRatio = screenWidth / screenHeight;
}

Camera::~Camera()
{

}

void Camera::SetupProjection( void )
{
    switch( projType ) {
    case ORTHOGRAPHIC:
        glViewport( 0, 0, this->screenWidth, this->screenHeight );
        gluOrtho2D( 0.0, this->screenWidth, 0.0, this->screenHeight );
        break;

    case PERSPECTIVE:
        glViewport( 0, 0, this->screenWidth, this->screenHeight );
        gluPerspective( this->fov, this->aspectRatio, this->nearPlane, this->farPlane );
        break;

    default:
        glLoadIdentity();
        break;
    }
}

void Camera::Resize( int screenWidth, int screenHeight )
{
    this->screenWidth = screenWidth;
    this->screenHeight = screenHeight;

    this->aspectRatio = screenWidth / screenHeight;
}

void Camera::MouseLook( Vec2f where, Vec2f last, int which )
{
    float rotateX = (float) ( where[ 0 ] - last[ 0 ] ) / this->screenWidth * M_PI;

    float rotateY = (float) ( where[ 1 ] - last[ 1 ] ) / this->screenHeight * M_PI;

    RotateView( -rotateY, cross( this->forwardVec, this->upVec ) );

    RotateView( -rotateX, Vec3f( 0.0, 1.0, 0.0 ) );
}

void Camera::RotateView( float angle )
{
    RotateView( angle, Vec3f( 0.0, 1.0, 0.0 ) );
}

void Camera::RotateView( float angle, Vec3f rotateVec )
{
    unitize( rotateVec );

    // compute a unit quaternion composing a rotation of angle about rotateVec
    Quat q( (float) sin( angle / 2 ) * rotateVec, cos( angle / 2 ) );
    Quat qInverse = inverse( q );

    // ( 0, p' ) = q * ( 0, p ) * q^-1
    this->forwardVec = ( q * Quat( this->forwardVec, 0 ) * qInverse ).vector();
    this->upVec = ( q * Quat( this->upVec, 0 ) * qInverse ).vector();

    unitize( this->forwardVec );
    unitize( this->upVec );
}


void Camera::Move( float speed )
{
    Move( speed, this->forwardVec );
}

void Camera::Move( float speed, Vec3f moveVec )
{
    unitize( moveVec );

    // move camera by speed in direction pointed to by moveVec
    this->pos += moveVec * speed;
}

void Camera::Look()
{
    // tell opengl where the camera is at and where it is looking
    gluLookAt( this->pos[ 0 ], this->pos[ 1 ], this->pos[ 2 ],
            this->pos[ 0 ] + this->forwardVec[ 0 ], this->pos[ 1 ] + this->forwardVec[ 1 ], this->pos[ 2 ] + this->forwardVec[ 2 ],
            this->upVec[ 0 ], this->upVec[ 1 ], this->upVec[ 2 ] );
}

#endif
