#include "fpscamera.h"

FPSCamera::FPSCamera( Vec3f pos, Vec3f forward_vec, Vec3f up_vec,
        float fov, float near_plane, float far_plane,
        int screen_width, int screen_height,
        ProjType proj_type ) :
    pos( pos ), forward_vec( forward_vec ), up_vec( up_vec ),
    fov( fov ), near_plane( near_plane ), far_plane( far_plane ),
    screen_width( screen_width ), screen_height( screen_height ),
proj_type( proj_type )
{
    this->rotate_speed = 0.05f;
    this->strafe_speed = 2.5f;
    this->move_speed   = 2.5f;

    this->aspect_ratio = screen_width / screen_height;
}

FPSCamera::~FPSCamera()
{

}

void FPSCamera::setup_projection( void )
{
    switch( proj_type )
    {
    case ORTHOGRAPHIC:
        glViewport( 0, 0, this->screen_width, this->screen_height );
        gluOrtho2D( 0.0, this->screen_width, 0.0, this->screen_height );
        break;

    case PERSPECTIVE:
        glViewport( 0, 0, this->screen_width, this->screen_height );
        gluPerspective( this->fov, this->aspect_ratio, this->near_plane, this->far_plane );
        break;

    default:
        glLoadIdentity();
        break;
    }
}

void FPSCamera::resize( int screen_width, int screen_height )
{
    this->screen_width = screen_width;
    this->screen_height = screen_height;

    this->aspect_ratio = screen_width / screen_height;
}

void FPSCamera::mouse_look( Vec2f where, Vec2f last, int which )
{
    float rotate_x = (float) ( where[ 0 ] - last[ 0 ] ) / this->screen_width * M_PI;

    float rotate_y = (float) ( where[ 1 ] - last[ 1 ] ) / this->screen_height * M_PI;

    rotate_view( -rotate_y, cross( this->forward_vec, this->up_vec ) );

    rotate_view( -rotate_x, Vec3f( 0.0, 1.0, 0.0 ) );
}

void FPSCamera::rotate_view( float angle )
{
    rotate_view( angle, Vec3f( 0.0, 1.0, 0.0 ) );
}

void FPSCamera::rotate_view( float angle, Vec3f rotate_vec )
{
    unitize( rotate_vec );

    // compute a unit quaternion composing a rotation of angle about rotate_vec
    Quat q( (float) sin( angle / 2 ) * rotate_vec, cos( angle / 2 ) );
    Quat q_inverse = inverse( q );

    // ( 0, p' ) = q * ( 0, p ) * q^-1
    this->forward_vec = ( q * Quat( this->forward_vec, 0 ) * q_inverse ).vector();
    this->up_vec = ( q * Quat( this->up_vec, 0 ) * q_inverse ).vector();

    unitize( this->forward_vec );
    unitize( this->up_vec );
}


void FPSCamera::move( float speed )
{
    move( speed, this->forward_vec );
}

void FPSCamera::move( float speed, Vec3f move_vec )
{
    unitize( move_vec );

    // move camera by speed in direction pointed to by move_vec
    this->pos += move_vec * speed;
}

void FPSCamera::look()
{
    // tell opengl where the camera is at and where it is looking
    gluLookAt( this->pos[ 0 ], this->pos[ 1 ], this->pos[ 2 ],  
            this->pos[ 0 ] + this->forward_vec[ 0 ], this->pos[ 1 ] + this->forward_vec[ 1 ], this->pos[ 2 ] + this->forward_vec[ 2 ],  
            this->up_vec[ 0 ], this->up_vec[ 1 ], this->up_vec[ 2 ] );
}

ostream & operator<<( ostream & out, const FPSCamera & camera )
{
    cout << "Position: " << camera.pos << endl;
    cout << "Forward:  " << camera.forward_vec << endl;
    cout << "Up:       " << camera.up_vec << endl;

    return out;
}

