#ifndef _CAMERA_H
#define _CAMERA_H

#include <gfx/gl.h>
#include <gfx/vec3.h>
#include <gfx/quat.h>

#include <iostream>
using namespace std;

// different types of projections the camera supports
enum ProjType
{
    ORTHOGRAPHIC,
    PERSPECTIVE
};

/*
 * a FPS'ish camera capable of rotating along all three cardinal
 * axes (extended from its current position, strafing along all 
 * three cardinal axes (again, extended from its current position), 
 * moving forwards and backwards, and free mouse-looking around its 
 * current position. Can be set to do either orthographic or perspective
 * projections with an arbitrary viewport and aspect ratio.
 */

class FPSCamera 
{

public:

    FPSCamera( Vec3f pos = Vec3f( 0.0f, 0.0f, 0.0f ), 
            Vec3f forward_vec = Vec3f( 1.0f, 0.0f, 1.0f ), 
            Vec3f up_vec = Vec3f( 0.0f, 1.0f, 0.0f ),
            float fov = 45.0f,
            float near_plane = 0.1f, float far_plane = 1000.0f,
            int screen_width = 640, int screen_height = 480,
            ProjType proj_type = PERSPECTIVE );

    ~FPSCamera();

    // bunch of getters and setters...      
    Vec3f get_position( void )                  {   return pos;                         }
    Vec3f get_forward_vector( void )                {   return forward_vec;                 }
    Vec3f get_up_vector( void )                 {   return up_vec;                      }

    float get_move_speed( void )                    {   return move_speed;                  }
    float get_strafe_speed( void )              {   return strafe_speed;                    }
    float get_rotate_speed( void )              {   return rotate_speed;                    }

    ProjType get_proj_type( void )              {   return proj_type;                   }
    float get_fov( void )                       {   return fov;                         }
    float get_aspect_ratio( void )              {   return aspect_ratio;                    }

    void set_position( Vec3f pos )              {   this->pos = pos;                    }
    void set_forward_vector( Vec3f forward_vec )    {   unitize( forward_vec ); this->forward_vec = forward_vec;        }
    void set_up_vector( Vec3f up_vec )              {   unitize( up_vec );      this->up_vec = up_vec;              }

    void set_move_speed( float move_speed )     {   this->move_speed = move_speed;      }
    void set_strafe_speed( float strafe_speed ) {   this->strafe_speed = strafe_speed;  }
    void set_rotate_speed( float rotate_speed ) {   this->rotate_speed = rotate_speed;  }

    void set_proj_type( ProjType proj_type )        {   this->proj_type = proj_type;            }
    void set_fov( float fov )                   {   this->fov = fov;                    }
    void set_aspect_ratio( float aspect_ratio ) {   this->aspect_ratio = aspect_ratio;  }

    // winow has been resized, update variables
    void resize( int screen_width, int screen_height );

    // rotate the camera's view "angle" radians around the specified axis
    virtual void rotate_view( float angle );
    virtual void rotate_view( float angle, Vec3f rotate_vec );

    // move the camera along the specified axis at the specified speed
    virtual void move( float speed );
    virtual void move( float speed, Vec3f move_vec );

    // use the mouse to freely look around the world
    virtual void mouse_look( Vec2f where, Vec2f last, int which );

    // tell OpenGL where to look using ook_at
    virtual void look( void );

    // setup the GL_PROJECTION matrix
    void setup_projection( void );

    // print out important variables
    friend ostream & operator<<( ostream & out, const FPSCamera & camera );

protected:

    // camera's position
    Vec3f pos;

    // camera's forward vector
    Vec3f forward_vec;                      

    // camera's up vector
    Vec3f up_vec;       

    // movement/rotation speeds
    float move_speed;
    float strafe_speed;
    float rotate_speed;

    // field of view
    float fov;

    // near and far clipping planes
    float near_plane, far_plane;

    // screen width and height for setting a viewport
    int screen_width, screen_height;

    // width / height, for setting perspective
    float aspect_ratio;

    // what type of camera are we
    ProjType proj_type;

};

#endif
