/**
 * @file msmodel.h
 * Header for MSModel class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _MSMODEL_H
#define _MSMODEL_H

#include "model.h"

#include <gfx/mat4.h>
#include <vector>

/**
 * A Milkshape3D graphical model.
 */
class MSModel : public Model {
public:
    /**
     * A triangle mesh, composed of a bunch of triangles, and a material.
     */
    struct Mesh {
        int material_idx;
        vector<int> triangle_idx;
    };

    /**
     * A triangle and its vertices' normal values, as well as texture
     * coords.
     */
    struct Triangle {
        Vec3f s, t;
        Vec3f normals[3];
        int vertex_idx[3];
    };

    /**
     * A single vertex and the bone it's attached to.
     */
    struct Vertex {
        char bone_idx;
        Vec3f pos;
    };

    /**
     * Keyframe for animation.
     */
    struct Keyframe {
        int bone_idx;
        float time; // sec
        Vec3f value;
    };

    /**
     * A skeletal bone, needed for skeletal animation.
     */
    struct Bone {
        Mat4 relative;
        Mat4 absolute;
        Mat4 current;   // the one that is applied when drawing

        vector<Keyframe> translate_frames;
        vector<Keyframe> rotate_frames;

        int cur_translate_frame;
        int cur_rotate_frame;

        int parent_idx;
    };

    MSModel( string dir, string filename );

    virtual ~MSModel();

    virtual void render();

    virtual void draw_selection()
    { draw_box( min, max, Vec3f(1.0f, 0.0f, 0.0f) ); }

    /**
     * Advance the animation by t seconds.
     * This involves interpolating between the two closest keyframes to our
     * current time, and updating our bones to the correct
     * location/orientation.
     */
    virtual void tick(double t);

    /**
     * Load in the textures from their files.
     */
    void reload_textures();

    /**
     * Load in the MS3D model from a model file.
     */
    bool load_model(string filename);

    Vec3 get_min() { return min; }
    Vec3 get_max() { return max; }

private:
    /**
     * Prepare the bone structures (only called once).
     * This fills in the bone matrices, and transforms all vertices in
     * the model to their own bone's space so animation changes them around
     * their bone.  Transformation back to world coordinates happens in
     * render().
     */
    void _setup_bones();

    /**
     * Reset the animation.
     */
    void _reset_animation();

    /**
     * @return the interpolation between the two keyframes we are between.
     */
    Vec3 _interp_keyframes( int curframe, const vector<Keyframe>& keyframes );

    /**
     * Adjust min,max values for our bbox based on the new vertex 'newpos'.
     * @param newpos a new vertex to consider for our bounding box.
     */
    void _compare_bbox( Vec3 newpos );
protected:
    string _dir;                    ///< the base dir for textures, etc
    vector<Material> _materials;    ///< list of materials used by the model
    vector<Mesh> _meshes;           ///< list of meshes that make up the model
    vector<Triangle> _triangles;    ///< list of triangles used in the mesh
    vector<Vertex> _vertices;       ///< list of vertices used in the mesh
    vector<Bone> _bones;            ///< list of bones in the skeleton
    double _time;                   ///< total animation time, sec
    double _curtime;                ///< current time
    Vec3 min;                       ///< minimum x,y,z (for bbox)
    Vec3 max;                       ///< maximum x,y,z (for bbox)
};

#endif // header guard
