#include "scene.h"

#include <gfx/gl.h>
#include <gfx/gltools.h>

static dWorldID _world_id;
static dSpaceID _space_id;
static dJointGroupID _contact_group;
static GameState* _gamestate;

dWorldID get_world_id( void ) { return _world_id; }
dSpaceID get_space_id( void ) { return _space_id; }

SceneManager::SceneManager( CL_GraphicContext* gc, GameState* gamestate ) :
    _selected_obj( NULL )
{
    _glstate = new CL_OpenGLState( gc );

    _gamestate = gamestate;

    gl_init();

    _current_level_id = -1;

    _world_id = dWorldCreate();

    // these should be set by the current level
    dWorldSetGravity( _world_id, 0.0f, -9.81f, 0.0f );
    dWorldSetERP( _world_id, 0.4f );
    dWorldSetCFM( _world_id, 1e-5f );

    _space_id = dSimpleSpaceCreate( 0 );

    _contact_group = dJointGroupCreate( 0 );
}

SceneManager::~SceneManager( void )
{
    dWorldDestroy( _world_id );
    dSpaceDestroy( _space_id );

    dJointGroupEmpty( _contact_group );
    dJointGroupDestroy( _contact_group );

    delete _glstate;
}

int SceneManager::register_signals( void )
{
    _slot_click = CL_Mouse::sig_key_down().connect(this, &SceneManager::on_mouse_down);
    _slot_move = CL_Mouse::sig_move().connect(this, &SceneManager::on_mouse_move);

    return 0;
}

void SceneManager::gl_init( void )
{
    _glstate->set_active();

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClearDepth(1.0);

    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

    glShadeModel(GL_SMOOTH);
    glEnable(GL_NORMALIZE);

    const Vec4f ambient_light(0.3f, 0.3f, 0.3f, 1.0f);
    const Vec4f light0_pos(5.0f, 10.0f, 5.0f, 0.0f);
    const Vec4f light0_rgb(1.0f, 1.0f, 1.0f, 1.0f);

    // Enable lighting and set up the lighting environment.  We specify a
    // global ambient glow and create two point lights.
    glEnable(GL_LIGHTING);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);

    glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_rgb);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_rgb);
    glEnable(GL_LIGHT0);

    glEnable(GL_TEXTURE_2D);
    
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    _camera.apply_projection();
}

void SceneManager::generate_handlers( void )
{
    RegisterEffect::iterator eit;

    Level::ObjectList objects = _gamestate->get_level()->get_objects();
    Level::ObjectIterator it;

    for ( it = objects.begin(); it != objects.end(); it++ ) {
        for ( eit = RegisterEffect::begin(); eit != RegisterEffect::end(); eit++ ) {
            (*eit)->add_handler( (*it).second );
        }
    }
}

void SceneManager::begin_level( void )
{
    if( _gamestate->get_current_state() == LEVEL_LOADING && 
        _current_level_id != _gamestate->get_level()->get_level_id() )
    {
        generate_handlers();

        _camera.set_radius( _gamestate->get_level()->_camera_radius );
        _camera.set_elevation( _gamestate->get_level()->_camera_elevation );
        _camera.set_azimuth( _gamestate->get_level()->_camera_azimuth );

        dJointGroupEmpty( _contact_group );
        
        _current_level_id = _gamestate->get_level()->get_level_id();
    }
}


void SceneManager::finish_level( void )
{
    if( _gamestate->get_current_state() == END_LEVEL && _current_level_id > 0 )
    {
        
        _current_level_id = -1;
    
    }
}

void SceneManager::near_callback(void* data, dGeomID o1, dGeomID o2)
{
    const unsigned long N = 6;
    dContact contacts[ N ];

    const unsigned long n = dCollide( o1, o2, N, &contacts[0].geom, sizeof(dContact) );

    for(unsigned long i = 0; i < n; ++i )
    {
        if( _gamestate->get_current_state() == RUN_SIM )
        {
            contacts[i].surface.mode = dContactBounce;
            contacts[i].surface.mu = 250;
            contacts[i].surface.bounce = 0.5f;
            contacts[i].surface.bounce_vel = 0.05f;
        }
        else if( _gamestate->get_current_state() == GAMEPLAY )
        {
            contacts[i].surface.mode = dContactBounce;
            contacts[i].surface.mu = 250;
        }

        dJointID c = dJointCreateContact( _world_id, _contact_group, &contacts[i] );

        dJointAttach( c, dGeomGetBody(contacts[i].geom.g1), dGeomGetBody(contacts[i].geom.g2) );
    }   
}


void SceneManager::tick(double t)
{
    begin_level();

    if( _gamestate->get_current_state() == RUN_SIM ) {
        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        Level::ActorList actors = _gamestate->get_level()->get_actors();
        Level::ActorIterator ait;

        // every actor and object should push any non-ode based (magnetism, 
        // wind, fear, etc.) effects for this tick into ode
        for ( ait = actors.begin(); ait != actors.end(); ait++ ) {
            (*ait).second->tick( t );
        }
        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {
            (*oit).second->tick( t );
        }

        dSpaceCollide( _space_id, 0, &SceneManager::near_callback );

        dWorldStep( _world_id, t );

        dJointGroupEmpty( _contact_group );
    }
    else if( _gamestate->get_current_state() == GAMEPLAY ) {
        /*
        dWorldSetGravity( _world_id, 0.0f, 0.0f, 0.0f );
        
        dSpaceCollide( _space_id, 0, &SceneManager::near_callback );

        dWorldStep( _world_id, t );

        dJointGroupEmpty( _contact_group );
        */

        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {
            (*oit).second->update_sim();
        }
    }

    finish_level();
}

void SceneManager::render()
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    if( _gamestate->get_current_state() == RUN_SIM || 
            _gamestate->get_current_state() == GAMEPLAY ||
            _gamestate->get_current_state() == LEVEL_LOADING )
    {
        _glstate->set_active();

        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        _camera.apply_view();

        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {
            glLoadName( (GLuint)(*oit).second );
            (*oit).second->render( ( (*oit).second == _selected_obj ) );
        }
    }
}

void SceneManager::on_mouse_down(const CL_InputEvent& event)
{
    if( _gamestate->get_current_state() == GAMEPLAY ) {
        _glstate->set_active();

        if ( event.id == CL_MOUSE_LEFT ) {
            int pos[] = {event.mouse_pos.x, event.mouse_pos.y};
            select_object(pos);
        }
        else if ( event.id == CL_MOUSE_RIGHT ) {

        }

        _prev_mouse_pos[0] = event.mouse_pos.x;
        _prev_mouse_pos[1] = event.mouse_pos.y;
    }
}

void SceneManager::on_mouse_move(const CL_InputEvent& event)
{
    if( _gamestate->get_current_state() == GAMEPLAY || 
        _gamestate->get_current_state() == RUN_SIM ) {
        Vec2 cur_mouse_pos( event.mouse_pos.x, event.mouse_pos.y );
        Object *obj = get_selected_object();

        if ( CL_Keyboard::get_keycode( CL_KEY_SHIFT ) ) {
            mouse_drag_camera( &_camera, _prev_mouse_pos, cur_mouse_pos );
        }
        else if ( obj && _gamestate->get_current_state() == GAMEPLAY ) {
            mouse_drag_object( &_camera, obj, _prev_mouse_pos, cur_mouse_pos );
        }

        _prev_mouse_pos[0] = event.mouse_pos.x;
        _prev_mouse_pos[1] = event.mouse_pos.y;
    }
}

void SceneManager::select_object(int *pos)
{
    GLuint buffer[128]; // may need to adjust the size

    _selected_obj = NULL;

    // let opengl know we want to pick an object
    begin_opengl_pick( pos, 1.0, buffer, 128 );

    // render the scene so opengl knows where things are
    _camera.apply_projection();
    render();

    // find out what object was picked
    GLuint select_idx = complete_opengl_pick( buffer );

    if (select_idx != opengl_pick_nil) {
        _selected_obj = (Object*)select_idx;
            //_gamestate->get_level()->get_objects()[select_idx];
    }
}
