#include "utils.h"

#include <math.h>
#include <gfx/gl.h>
#include <stdio.h>

// Milkshape3D likes to store their rotations as (rx, ry, rz) triples,
// where each is an angle to rotate around that axis.
Mat4 rotation_matrix_angles_rad( const Vec3& angles )
{
    double cr = cos( angles[ 0 ] );
    double sr = sin( angles[ 0 ] );
    double cp = cos( angles[ 1 ] );
    double sp = sin( angles[ 1 ] );
    double cy = cos( angles[ 2 ] ); 
    double sy = sin( angles[ 2 ] );
    double srsp = sr * sp;
    double crsp = cr * sp;

    Mat4 mat;

    mat( 0, 0 ) = cp * cy;
    mat( 1, 0 ) = cp * sy;
    mat( 2, 0 ) = -sp;

    mat( 0, 1 ) = srsp * cy - cr * sy;
    mat( 1, 1 ) = srsp * sy + cr * cy;
    mat( 2, 1 ) = sr * cp;

    mat( 0, 2 ) = crsp * cy + sr * sy;
    mat( 1, 2 ) = crsp * sy - sr * cy;
    mat( 2, 2 ) = cr * cp;

    mat( 3, 3 ) = 1.0;

    return mat;
}

// useful for pushing an object's rotation onto the modelview_matrix stack
void dMatrix3ToOpenGL( float oglMat[ 16 ], const dMatrix3 dMat3 )
{
    oglMat[  0 ] = dMat3[  0 ];
    oglMat[  1 ] = dMat3[  4 ];
    oglMat[  2 ] = dMat3[  8 ];
    oglMat[  3 ] = 0.0f; 

    oglMat[  4 ] = dMat3[  1 ];
    oglMat[  5 ] = dMat3[  5 ];
    oglMat[  6 ] = dMat3[  9 ];
    oglMat[  7 ] = 0.0f;

    oglMat[  8 ] = dMat3[  2 ];
    oglMat[  9 ] = dMat3[  6 ];
    oglMat[ 10 ] = dMat3[ 10 ];
    oglMat[ 11 ] = 0.0f;

    oglMat[ 12 ] = 0.0f;
    oglMat[ 13 ] = 0.0f;
    oglMat[ 14 ] = 0.0f;
    oglMat[ 15 ] = 1.0f;
}

void Mat4TodMatrix3( dMatrix3 dmat3, const Mat4 mat4 )
{
    dmat3[  0 ] = mat4[ 0 ][ 0 ];
    dmat3[  1 ] = mat4[ 1 ][ 0 ];
    dmat3[  2 ] = mat4[ 2 ][ 0 ];
    dmat3[  3 ] = mat4[ 3 ][ 0 ];
    
    dmat3[  4 ] = mat4[ 0 ][ 1 ];
    dmat3[  5 ] = mat4[ 1 ][ 1 ];
    dmat3[  6 ] = mat4[ 2 ][ 1 ];
    dmat3[  7 ] = mat4[ 3 ][ 1 ];
    
    dmat3[  8 ] = mat4[ 0 ][ 2 ];
    dmat3[  9 ] = mat4[ 1 ][ 2 ];
    dmat3[ 10 ] = mat4[ 2 ][ 2 ];
    dmat3[ 11 ] = mat4[ 3 ][ 2 ];
}

void dMatrix3ToMat4( Mat4 mat4, const dMatrix3 dmat3 )
{
    mat4[ 0 ][ 0 ] = dmat3[  0 ];
    mat4[ 0 ][ 1 ] = dmat3[  4 ];
    mat4[ 0 ][ 2 ] = dmat3[  8 ];
    mat4[ 0 ][ 3 ] = 0.0;

    mat4[ 1 ][ 0 ] = dmat3[  1 ];
    mat4[ 1 ][ 1 ] = dmat3[  5 ];
    mat4[ 1 ][ 2 ] = dmat3[  9 ];
    mat4[ 1 ][ 3 ] = 0.0;

    mat4[ 2 ][ 0 ] = dmat3[  2 ];
    mat4[ 2 ][ 1 ] = dmat3[  6 ];
    mat4[ 2 ][ 2 ] = dmat3[ 10 ];
    mat4[ 2 ][ 3 ] = 0.0;

    mat4[ 3 ][ 0 ] = dmat3[  3 ];
    mat4[ 3 ][ 1 ] = dmat3[  7 ];
    mat4[ 3 ][ 2 ] = dmat3[ 11 ];
    mat4[ 3 ][ 3 ] = 1.0;
}

void dMatrix3TodMatrix3( dMatrix3 dmat3_1, const dMatrix3 dmat3_2 )
{
    dmat3_1[  0 ] = dmat3_2[  0 ];
    dmat3_1[  1 ] = dmat3_2[  1 ];
    dmat3_1[  2 ] = dmat3_2[  2 ];
    dmat3_1[  3 ] = dmat3_2[  3 ];
    dmat3_1[  4 ] = dmat3_2[  4 ];
    dmat3_1[  5 ] = dmat3_2[  5 ];
    dmat3_1[  6 ] = dmat3_2[  6 ];
    dmat3_1[  7 ] = dmat3_2[  7 ];
    dmat3_1[  8 ] = dmat3_2[  8 ];
    dmat3_1[  9 ] = dmat3_2[  9 ];
    dmat3_1[ 10 ] = dmat3_2[ 10 ];
    dmat3_1[ 11 ] = dmat3_2[ 11 ];
}

void dRealPtrTodMatrix4( dMatrix4 dmat4, const float* fltPtr )
{
    dmat4[  0 ] = fltPtr[  0 ];
    dmat4[  1 ] = fltPtr[  1 ];
    dmat4[  2 ] = fltPtr[  2 ];
    dmat4[  3 ] = 0.0f;

    dmat4[  4 ] = fltPtr[  3 ];
    dmat4[  5 ] = fltPtr[  4 ];
    dmat4[  6 ] = fltPtr[  5 ];
    dmat4[  7 ] = 0.0f;

    dmat4[  8 ] = fltPtr[  6 ];
    dmat4[  9 ] = fltPtr[  7 ];
    dmat4[ 10 ] = fltPtr[  8 ];
    dmat4[ 11 ] = 0.0f;

    dmat4[ 12 ] = fltPtr[  9 ];
    dmat4[ 13 ] = fltPtr[ 10 ];
    dmat4[ 14 ] = fltPtr[ 11 ];
    dmat4[ 15 ] = 0.0f;
}

void dRealPtrTodMatrix3( dMatrix3 dmat3, const float* fltPtr )
{
    dmat3[  0 ] = fltPtr[  0 ];
    dmat3[  1 ] = fltPtr[  1 ];
    dmat3[  2 ] = fltPtr[  2 ];
    dmat3[  3 ] = fltPtr[  3 ];
    dmat3[  4 ] = fltPtr[  4 ];
    dmat3[  5 ] = fltPtr[  5 ];
    dmat3[  6 ] = fltPtr[  6 ];
    dmat3[  7 ] = fltPtr[  7 ];
    dmat3[  8 ] = fltPtr[  8 ];
    dmat3[  9 ] = fltPtr[  9 ];
    dmat3[ 10 ] = fltPtr[ 10 ];
    dmat3[ 11 ] = fltPtr[ 11 ];
}

void dRealPtrTodQuaternion( dQuaternion dquat, const float* fltPtr )
{
    dquat[ 0 ] = fltPtr[ 0 ];
    dquat[ 1 ] = fltPtr[ 1 ];
    dquat[ 2 ] = fltPtr[ 2 ];
    dquat[ 3 ] = fltPtr[ 3 ];
}

void dRealPtrToVec3( Vec3& vec3, const float* fltPtr )
{
    vec3[ 0 ] = fltPtr[ 0 ];
    vec3[ 1 ] = fltPtr[ 1 ];
    vec3[ 2 ] = fltPtr[ 2 ];
}

/**
 * Set the translation vectors for a transformation matrix.
 */
void set_translation( Mat4& mat, const Vec3& translate )
{
    mat( 0, 3 ) = translate[ 0 ];
    mat( 1, 3 ) = translate[ 1 ];
    mat( 2, 3 ) = translate[ 2 ];
}

/**
 * Invert a translation*rotation matrix.
 * Optimized inverse for T*R matrix.
 */
Mat4 inverse_t_r( const Mat4& mat )
{
    // T = translation matrix, R = rotation matrix, tv = T.col(3); 
    // (TR)^-1 = R^-1 * T^-1 = Rt * T^-1 = [Rt, -tv]
    Vec3 translate = proj( mat.col( 3 ) );

    Mat4 inv = transpose( mat );
    inv[ 3 ] = Vec4( 0.0, 0.0, 0.0, 1.0 );
    set_translation( inv, inv * -translate );

    return inv;
}

struct textureImage {
    int width;
    int height;
    unsigned char *data;
};

/* simple loader for 24bit bitmaps (data is in rgb-format) */
int load_bmp( const char *filename, textureImage *texture )
{
    FILE *file;
    unsigned short int bfType;
    long int bfOffBits;
    short int biPlanes;
    short int biBitCount;
    long int biSizeImage;
    int i;
    unsigned char temp;
    /* make sure the file is there and open it read-only (binary) */
    if ((file = fopen(filename, "rb")) == NULL)
    {
        printf("File not found : %s\n", filename);
        return 0;
    }
    if(!fread(&bfType, sizeof(short int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    /* check if file is a bitmap */
    if (bfType != 19778)
    {
        printf("Not a Bitmap-File!\n");
        return 0;
    }        
    /* get the file size */
    /* skip file size and reserved fields of bitmap file header */
    fseek(file, 8, SEEK_CUR);
    /* get the position of the actual bitmap data */
    if (!fread(&bfOffBits, sizeof(long int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    printf("Data at Offset: %ld\n", bfOffBits);
    /* skip size of bitmap info header */
    fseek(file, 4, SEEK_CUR);
    /* get the width of the bitmap */
    fread(&texture->width, sizeof(int), 1, file);
    printf("Width of Bitmap: %d\n", texture->width);
    /* get the height of the bitmap */
    fread(&texture->height, sizeof(int), 1, file);
    printf("Height of Bitmap: %d\n", texture->height);
    /* get the number of planes (must be set to 1) */
    fread(&biPlanes, sizeof(short int), 1, file);
    if (biPlanes != 1)
    {
        printf("Error: number of Planes not 1!\n");
        return 0;
    }
    /* get the number of bits per pixel */
    if (!fread(&biBitCount, sizeof(short int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    printf("Bits per Pixel: %d\n", biBitCount);
    if (biBitCount != 24)
    {
        printf("Bits per Pixel not 24\n");
        return 0;
    }
    /* calculate the size of the image in bytes */
    biSizeImage = texture->width * texture->height * 3;
    printf("Size of the image data: %ld\n", biSizeImage);
    texture->data = (unsigned char *)malloc(biSizeImage);
    /* seek to the actual data */
    fseek(file, bfOffBits, SEEK_SET);
    if (!fread(texture->data, biSizeImage, 1, file))
    {
        printf("Error loading file!\n");
        return 0;
    }
    /* swap red and blue (bgr -> rgb) */
    for (i = 0; i < biSizeImage; i += 3)
    {
        temp = texture->data[i];
        texture->data[i] = texture->data[i + 2];
        texture->data[i + 2] = temp;
    }
    return 1;
}

uint load_gl_texture( string filename )
{
    bool status;
    textureImage *texti;
    GLuint texPntr[1];
    
    status = false;
    texti = (textureImage *)malloc(sizeof(textureImage));
    if (load_bmp(filename.c_str(), texti))
    {
        status = true;
        glGenTextures(1, &texPntr[0]);   /* create the texture */
        glBindTexture(GL_TEXTURE_2D, texPntr[0]);
        /* actually generate the texture */
        glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0,
            GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        /* enable linear filtering */
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    /* free the ram we used in our texture generation process */
    if (texti)
    {
        if (texti->data)
            free(texti->data);
        free(texti);
    }    
    return texPntr[0];
}
